/*
 * Copyright (C) 2019-2023 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/dma-mapping.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <linux/list.h>
#include <linux/hrtimer.h>
#include <linux/sched.h>

#include "../../include/fw-board.h"
#include "fw-drv.h"
#include "fw-pci.h"

#define DRV_NAME "fipwatcher"
#define BOARD_INFO "FIPWatcher Device"

/* Module parameters */
static unsigned int max_devices = MAX_FW_BOARD;

/* Module globals */
static LIST_HEAD(fw_list); /* sorted by ascending minor number */
static struct class *fw_class;
static dev_t fw_dev_first;

int pci_major = PCI_MAJOR;
static uint8_t minors[MAX_FW_BOARD] = {0};

static int dma_xfer(struct fw_dev *fw, int direction, void *host_buffer,
    uint32_t target_addr, size_t bsz);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
    static irqreturn_t fw_irq_handler(int irq, void *dev_id)
#else
    static irqreturn_t fw_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    struct fw_dev *fw = (struct fw_dev *) dev_id;
    int irq_number = irq - fw->infos.irq_number;
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_INTCSR_REG);
    uint32_t reg_data = ioread32(reg_addr);

    if (!fw->has_msi_cap && irq_number) {
        /**
         * For a Line-Based interrupt, irq_number should be always equals to 0
         * => irq not produced by this device
         */
        return IRQ_NONE;
    }

    /* is the IRQ signal is produced by the device ? */
    if ((reg_data & PLX_INTCSR_LINT1_ACT) ||
        (reg_data & PLX_INTCSR_LINT2_ACT)) {

        spin_lock(&fw->it_lock);

        /* irq produced by this device */
        /* disable IRQ */
        fw_irq_global_enable(fw, 0);

        fw->irq.src = 0;
        if (reg_data & PLX_INTCSR_LINT1_ACT)
            fw->irq.src |= FW_IRQ_LINT1;

        if (reg_data & PLX_INTCSR_LINT2_ACT)
            fw->irq.src |= FW_IRQ_LINT2;

        fw->irq.cnt += 1;

        /* raise the interrupt flag + wake up the mutex */
        fw->it_flag = 1;
        spin_unlock(&fw->it_lock);
        wake_up(&fw->it_mutex);

        return IRQ_HANDLED;
    } else {
        /* irq not produced by this device */
        return IRQ_NONE;
    }
}

static void fw_register(struct fw_dev *fw, int index)
{
    char dev_name[64];
    int err, devno = MKDEV(pci_major, index);

    /* add class to device - ls /dev/ */
    sprintf(dev_name, "%s%d", DRV_NAME, index);
    fw->dev = device_create(fw_class, NULL, devno, NULL, dev_name);
    if (IS_ERR(fw->dev))
        printk(KERN_ERR DRV_NAME
            ": Error when adding class to device (index %d)", index);

    /* add char device */
    cdev_init(&fw->cdev, &fw_fops);
    fw->cdev.owner = THIS_MODULE;
    fw->cdev.ops = &fw_fops;
    err = cdev_add(&fw->cdev, devno, max_devices);
    if (err)
        printk(KERN_ERR DRV_NAME
            ": Error %d when adding char device (index %d)", err, index);
}

static void fw_unregister(struct fw_dev *fw)
{
    /* device destroying */
    if (fw_class) {
        int devno = MKDEV(pci_major, fw->minor);
        device_destroy(fw_class, devno);
    }
    cdev_del(&fw->cdev);
}

static int fw_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
    struct fw_dev *fw;
    uint8_t irq_pin;
    uint8_t bar_mem_cnt;
    uint8_t free_minor_dectected = 0;
    unsigned long flag;
    int err, i, nb_vecs = 0, msi_pos = 0;

    spinlock_t to_lock = __SPIN_LOCK_UNLOCKED(to_lock);

    printk(KERN_DEBUG DRV_NAME ": ==> Device probing\n");

    fw = kcalloc(1, sizeof(struct fw_dev), GFP_KERNEL);
    if (!fw)
        return -ENODEV;

    err = pci_enable_device(pdev);
    if (err)
        return err;

    /* Enable: MSI + DMA support */
    pci_set_master(pdev);

    /* MSI capabilities ? */
    if ((msi_pos = pci_find_capability(pdev, PCI_CAP_ID_MSI)) > 0)
        fw->has_msi_cap = true;

    fw->pdev = pdev;

    spin_lock_init(&to_lock);

    spin_lock(&to_lock);
    for (i = 0; i < MAX_FW_BOARD; i++) {
        if (minors[i] == 0) {
            minors[i] = 1 ;
            fw->minor = i;
            free_minor_dectected = 1;
            break;
        }
    }

    if (free_minor_dectected == 0)
        return -EFAULT;

    spin_unlock(&to_lock);
    pci_set_drvdata(pdev, fw);
    fw_register(fw, fw->minor);

    /* Reserve PCI I/O and memory resources */
    err = pci_request_regions(pdev, DRV_NAME);
    if (err)
        return -ENODEV;

    bar_mem_cnt = 0;
    for (i = 0; i < DEVICE_COUNT_RESOURCE; i++)  {

        if (pci_resource_start(pdev, i) != 0) {
            printk(KERN_DEBUG DRV_NAME 
                ": BAR %d (%#08x - %#08x), len = %d, flags = %#08x\n", i,
                (uint32_t) pci_resource_start(pdev, i),
                (uint32_t) pci_resource_end(pdev, i),
                (uint32_t) pci_resource_len(pdev, i),
                (uint32_t) pci_resource_flags(pdev, i));
        }

        flag = pci_resource_flags(pdev, i);

        if ((flag & IORESOURCE_MEM) == IORESOURCE_MEM) {
            fw->infos.bar_bsz[bar_mem_cnt] = pci_resource_len(pdev, i);
            fw->infos.bar_base[bar_mem_cnt] = pci_resource_start(pdev, i);
            fw->remap[bar_mem_cnt] = ioremap(pci_resource_start(pdev, i),
                                            pci_resource_len(pdev, i));
            if (fw->remap[bar_mem_cnt] == NULL)
                return -ENOMEM;

            bar_mem_cnt++;
        }
    }

    fw->it_flag = 0;
    init_waitqueue_head (&fw->it_mutex);
    spin_lock_init(&fw->it_lock);

    err = pci_read_config_byte(pdev, PCI_INTERRUPT_PIN, &irq_pin);
    if (err)
        return err;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    nb_vecs = pci_alloc_irq_vectors(pdev, 1, IRQ_VECS_MAX,
                PCI_IRQ_LEGACY | PCI_IRQ_MSI);
#else
    nb_vecs = pci_enable_msi_range(pdev, 1, IRQ_VECS_MAX);
#endif

    if (nb_vecs < 0)
        return -ENOSPC;

    for (i = 0; i < nb_vecs; i++)  {
        err = request_irq(pdev->irq+i, &fw_irq_handler,
                IRQF_SHARED, DRV_NAME, fw);
        if (err)
            return err;
    }

    init_interrupt(fw);

    fw->nb_irq_vecs = nb_vecs;
    fw->infos.bar_cnt = bar_mem_cnt;
    err = fw_board_init(fw);
    if (err)
        return err;

    /* MSI capabilities registration */
    if (fw->has_msi_cap) {
        uint32_t msi_addr_lo = 0;
        uint32_t msi_addr_hi = 0;
        pci_read_config_dword(pdev, msi_pos + PCI_MSI_ADDRESS_LO, &msi_addr_lo);
        pci_read_config_dword(pdev, msi_pos + PCI_MSI_ADDRESS_HI, &msi_addr_hi);
        iowrite32(msi_addr_lo, fw->remap[BAR_BOARD] + MSI_ADDR_LO_REG);
        iowrite32(msi_addr_hi, fw->remap[BAR_BOARD] + MSI_ADDR_HI_REG);
        /* TODO: case of msi vectors > 4 (not supported here) */
    }

    printk(KERN_DEBUG DRV_NAME ": DMA Capability = %s\n",
        (fw->has_dma_cap) ? "YES" : "NO");
    printk(KERN_DEBUG DRV_NAME ": MSI Capability = %s\n",
        (fw->has_msi_cap) ? "YES" : "NO");
    printk(KERN_DEBUG DRV_NAME
        ": IRQ pin  = #%d (0=none, 1=INTA#...4=INTD#)\n", irq_pin);
    printk(KERN_DEBUG DRV_NAME ": IRQ line = #%d\n", pdev->irq);
    printk(KERN_DEBUG DRV_NAME ": IRQ vectors = %d\n", nb_vecs);

    spin_lock(&to_lock);
    list_add_tail(&fw->link, &fw_list);
    spin_unlock(&to_lock);

    return 0;
}

static void fw_remove(struct pci_dev *pdev)
{
    spinlock_t to_lock = __SPIN_LOCK_UNLOCKED(to_lock);
    struct fw_dev *fw = pci_get_drvdata(pdev);
    int i = 0;

    printk(KERN_DEBUG DRV_NAME ": ==> Device removing (idx=%d)\n",
           fw->minor);

    pci_clear_master(pdev);

    /* free IRQ */
    if (fw->infos.irq_number >= 0) {
        printk(KERN_DEBUG DRV_NAME
            ": Freeing IRQ #%d\n", fw->infos.irq_number);
        /* disable irq */
        for (i = 0; i < fw->nb_irq_vecs; i++)
            free_irq(pdev->irq+i, (void *) fw);
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    pci_free_irq_vectors(pdev);
#else
    pci_disable_msi(pdev);
#endif

    /* iounmap safe for NULL */
    for (i = 0; i < fw->infos.bar_cnt; ++i)
        iounmap(fw->remap[i]);

    spin_lock_init(&to_lock);

    spin_lock(&to_lock);
    minors[fw->minor] = 0;
    spin_unlock(&to_lock);

    pci_disable_device(pdev);
    pci_release_regions(pdev);

    fw_unregister(fw);

    spin_lock(&to_lock);
    list_del(&fw->link);
    spin_unlock(&to_lock);

    kfree(fw->dma_buffer);
    kfree(fw);
}

int fw_open(struct inode *inode, struct file *file)
{
    int minor = MINOR(inode->i_rdev);
    struct list_head *cur;
    struct fw_dev *fw;

    list_for_each(cur, &fw_list) {

        fw = list_entry(cur, struct fw_dev, link);
        if (fw->minor == minor) {

            if (fw->is_open) {
                printk(KERN_WARNING DRV_NAME
                    ": Device is already open (index %d)\n",
                    fw->minor);
                return -EBUSY;
            }

            init_interrupt(fw);
            file->private_data = fw;

            fw->is_open = 1;

            return 0;
        }
    }

    return -ENODEV;
}

int fw_release(struct inode *inode, struct file *file)
{
    struct fw_dev *fw;

    if (file == NULL)
        return 0;

    if (file->private_data == NULL)
        return 0;

    fw = file->private_data;
    fw->is_open = 0;
    file->private_data = 0;

    return 0;
}

static int dma_xfer_wait_for_end(struct fw_dev *fw, uint32_t timeout_us)
{
    int err;
    void *dma_base = fw->remap[BAR_BOARD] + DMA_AREA_PCI_OFFSET;
    uint32_t reg_val;

    /* polling mode */
    err = wait_for_reg32(dma_base + DMA_STA_REG(0), DMA_STA_COMPL_OK,
            timeout_us, 1);
    if (!err)
        return 0;

    /* dma error trace */
    reg_val = ioread32(dma_base + DMA_STA_REG(0));
    printk(KERN_ERR DRV_NAME
        ": dma_xfer_wait_for_end => dma_status_reg=0x%08x (minor=%d)\n",
        reg_val, fw->minor);
    return err;
}

static int dma_xfer(struct fw_dev *fw, int direction, void *host_buffer,
        uint32_t target_addr, size_t bsz)
{
    int err = 0;
    void *dma_base = fw->remap[BAR_BOARD] + DMA_AREA_PCI_OFFSET;
    uint32_t data;
    dma_addr_t bus_addr;

    if (!host_buffer)
        return -EACCES;

    if (bsz > DMA_BUFFER_BSZ)
        return -EACCES;

    if (fw->dma_xfer_in_progress)
        return -EBUSY;

    /* map the buffer for dma */
    fw->dma_dir = ((direction == DMA_OP_W) ? \
                            DMA_TO_DEVICE : DMA_FROM_DEVICE);
    fw->dma_bsz = bsz;
    bus_addr = dma_map_single(&fw->pdev->dev, fw->dma_buffer,
                fw->dma_bsz, fw->dma_dir);
    if (dma_mapping_error(&fw->pdev->dev, bus_addr))
        return -EACCES;

    fw->dma_addr = bus_addr;

    /* set up the device */
    /* => set source parameter */
    data = (direction == DMA_OP_R) ? DMA_SRC_PARAM_OP_R : 0;
    iowrite32(data, dma_base + DMA_SRC_PARAM_REG(0));
    /* => set destination parameter */
    iowrite32(0, dma_base + DMA_DST_PARAM_REG(0));
    /* => set source address */
    data = (uint32_t) ((direction == DMA_OP_W) ? bus_addr : target_addr);
    iowrite32(data, dma_base + DMA_SRC_ADDR_LO_REG(0));
    iowrite32(0, dma_base + DMA_SRC_ADDR_HI_REG(0));
    /* => set destination address */
    data = (uint32_t) ((direction == DMA_OP_W) ? target_addr : bus_addr);
    iowrite32(data, dma_base + DMA_DST_ADDR_LO_REG(0));
    iowrite32(0, dma_base + DMA_DST_ADDR_HI_REG(0));
    /* => set length */
    iowrite32(fw->dma_bsz, dma_base + DMA_LEN_REG(0));

    /* prepare dma buffer (for write)*/
    if (direction == DMA_OP_W)
        memcpy(fw->dma_buffer, host_buffer, fw->dma_bsz);

    /* start the operation */
    fw->dma_xfer_in_progress = true;
    data = DMA_CTRL_OP_STRT_ABRT;
    iowrite32(data, dma_base + DMA_CTRL_REG(0));

    err = dma_xfer_wait_for_end(fw, DMA_TMO_US);
    fw->dma_xfer_in_progress = false;

    /* retrieve dma buffer (for read) */
    if ((!err) && (direction == DMA_OP_R))
        memcpy(host_buffer, fw->dma_buffer, fw->dma_bsz);

    dma_unmap_single(&fw->pdev->dev, bus_addr, fw->dma_bsz,
        fw->dma_dir);

    return err;
}

static int fw_info_get(struct fw_dev *fw, void __user *uarg)
{
    return (copy_to_user(uarg, &fw->infos, sizeof(struct fw_dev_infos))) ? \
            -EFAULT : 0;
}

static int fw_act_set(struct fw_dev *fw, void __user *uarg)
{
    struct fw_act_set_cmd cmd;
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);

    memset(&cmd, 0, sizeof(cmd));
    if (copy_from_user(&cmd, uarg, sizeof(cmd)))
        return -EFAULT;

    if (cmd.op)
        reg_data |= CTRL_CTL_FIP_ENA;
    else
        reg_data &= ~CTRL_CTL_FIP_ENA;

    /* clear it (to get a clean context) */
    reg_data |= CTRL_CTL_CLR_INT;

    iowrite32(reg_data, reg_addr);
    return 0;
}

static int fw_epoch_set(struct fw_dev *fw, void __user *uarg)
{
    struct fw_epoch_set_cmd cmd;
    uint32_t *reg_lo_addr = (uint32_t *) (fw->remap[BAR_BOARD] + EPOCH_LO_REG);
    uint32_t *reg_hi_addr = (uint32_t *) (fw->remap[BAR_BOARD] + EPOCH_HI_REG);
    uint32_t *reg_ctl_addr = (uint32_t *) (fw->remap[BAR_BOARD] + CTRL_REG);
    uint32_t reg_ctl_data = ioread32(reg_ctl_addr);

    memset(&cmd, 0, sizeof(cmd));
    if (copy_from_user(&cmd, uarg, sizeof(cmd)))
        return -EFAULT;

    /* set new epoch value */
    iowrite32((uint32_t) cmd.epoch, reg_lo_addr);
    iowrite32((uint32_t) (cmd.epoch >> 32), reg_hi_addr);

    /* set epoch flag */
    reg_ctl_data |= CTRL_CTL_SET_EPOCH;
    iowrite32(reg_ctl_data, reg_ctl_addr);
    wait_for_us(10);
    reg_ctl_data &= ~CTRL_CTL_SET_EPOCH;
    iowrite32(reg_ctl_data, reg_ctl_addr);
    return 0;
}

static int fw_irq_wait_for(struct fw_dev *fw, void __user *uarg)
{
    struct arg_io io;
    int err = 0;

    if (copy_from_user(&io, uarg, sizeof(io)))
        return -EFAULT;

    if (!io.o_ptr)
        return -EINVAL;

    if (io.o_bsz < sizeof(struct fw_dev_irq))
        return -EINVAL;

    /* irq enable */
    fw_irq_global_enable(fw, 1);

    if (!wait_event_interruptible_timeout( \
            fw->it_mutex,
            fw->it_flag,
            msecs_to_jiffies(IRQ_TIMEOUT_MS)))
    {
        return -ETIME;
    }

    /* drop the interrupt flag + wake up the mutex */
    spin_lock(&fw->it_lock);
    fw->it_flag = 0;

#if !defined(FW_VIA_MMAP)
    {
        uint32_t i, ctl_data, wsz = FW_MEM_WSZ;
        uint32_t *fw_buf_addr = (uint32_t *) (fw->remap[BAR_FW] + FW_MEM_OFFSET);
        uint32_t *ctl_addr = (uint32_t *) (fw->remap[BAR_BOARD] + CTRL_REG);
        if (fw->has_dma_cap)
        {
            /* dma reading */
            err = dma_xfer(fw, DMA_OP_R, fw->irq.data, FW_MEM_OFFSET, FW_MEM_BSZ);
        }
        else
        {
            /* mmio reading */
            for (i = 0; i < wsz; ++i)
                *(((uint32_t*) fw->irq.data) + i) = ioread32(fw_buf_addr + i);
        }

        /* clear bar mem */
        for (i = 0; i < wsz; ++i)
            iowrite32(0, fw_buf_addr + i);

        /**
         * Note: the useful data has been retrieved,
         * now we can clear the interrupt.
         */
        ctl_data = ioread32(ctl_addr);
        ctl_data |= CTRL_CTL_CLR_INT;
        iowrite32(ctl_data, ctl_addr);
    }
#endif

    spin_unlock(&fw->it_lock);

    if (err)
        return err;

    /* copy irq data to user space */
    if (copy_to_user((void *) (uintptr_t) io.o_ptr, &fw->irq,
        sizeof(struct fw_dev_irq)))
        return -EFAULT;

    /* update arg_io */
    io.status = sizeof(struct fw_dev_irq);
    return (copy_to_user(uarg, &io, sizeof(io))) ? -EFAULT : 0;
}

static int fw_irq_resume(struct fw_dev *fw, void __user *uarg)
{
    /**
     * Wake up 'manually' the mutex
     * Note: Useful to unlock a wait on an event that won't come
     */
    spin_lock(&fw->it_lock);
    /* disable irq */
    fw_irq_global_enable(fw, 0);
    fw->it_flag = 1;
    spin_unlock(&fw->it_lock);
    wake_up(&fw->it_mutex);
    return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35))
    int fw_ioctl(struct inode *inode, struct file *file,
        unsigned int cmd, unsigned long arg)
#else
    long fw_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#endif
{
    struct fw_dev *fw = file->private_data;
    void __user *uarg = (void __user *) arg;
    int err = 0;

    /* check command */
    if (_IOC_TYPE(cmd) != FW_IOCTL_MAGIC)
        return -ENOTTY;

    if (_IOC_DIR(cmd) & _IOC_READ) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        err = !access_ok(uarg, _IOC_SIZE(cmd));
#else
        err = !access_ok(VERIFY_WRITE, uarg, _IOC_SIZE(cmd));
#endif
    }

    if (_IOC_DIR(cmd) & _IOC_WRITE) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
        err = !access_ok(uarg, _IOC_SIZE(cmd));
#else
        err = !access_ok(VERIFY_READ, uarg, _IOC_SIZE(cmd));
#endif
    }
    if (err)
        return -EFAULT;

    switch (cmd) {
    case FW_IOCTL_INFO_GET:
        err = fw_info_get(fw, uarg);
        break;
    case FW_IOCTL_ACT_SET:
        err = fw_act_set(fw, uarg);
        break;
    case FW_IOCTL_EPOCH_SET:
        err = fw_epoch_set(fw, uarg);
        break;
    case FW_IOCTL_RESET:
        err = fw_reset(fw, uarg);
        break;
    case FW_IOCTL_IRQ_WAIT_FOR:
        err = fw_irq_wait_for(fw, uarg);
        break;
    case FW_IOCTL_IRQ_RESUME:
        err = fw_irq_resume(fw, uarg);
        break;
    default:
        printk(KERN_ERR DRV_NAME ": Unsupported IOCTL command (0x%x)\n", cmd);
        return -EINVAL;
    }
    return err;
}

struct file_operations fw_fops = {
    .owner = THIS_MODULE,
    .open  = fw_open,
    .release = fw_release,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35))
    .ioctl = fw_ioctl,
#else
    .unlocked_ioctl = fw_ioctl,
    .compat_ioctl = fw_ioctl,
#endif
};

/* PLX */
// For backward compatibility but these values are not normal.
// Values should be:
// PLX 9052: Vendor ID: 0x10b5 (PLX) - Device ID: 0x9052
// PLX 9030: Vendor ID: 0x10b5 (PLX) - Device ID: 0x9030
#define PLX_VID 0x10b5
#define PLX_DID 0x5201
/* Microsemi */
#define ACTEL_VID 0x11aa
#define ACTEL_DID 0x1556

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
static struct pci_device_id fw_idtable[] = {
    /* backward compatibility */
    { PCI_DEVICE_SUB(PLX_VID, PLX_DID, PCI_ANY_ID, FW_SSDID_MPCIE_COP) },
    /* current compatibility */
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCI_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCIE_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCI104_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCIE104_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PXI_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PXIE_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_MPCIE_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_CPCI_COP) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCI_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCIE_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCI104_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PCIE104_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PXI_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_PXIE_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_MPCIE_OPT) },
    { PCI_DEVICE_SUB(ACTEL_VID, ACTEL_DID, PCI_ANY_ID, FW_SSDID_CPCI_OPT) },
    { 0, }
};
#else
static const struct pci_device_id fw_idtable[] = {
    /* backward compatibility */
    { .vendor = PLX_VID, .device = PLX_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_MPCIE_COP, },
    /* current compatibility */
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCI_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCIE_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCI104_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCIE104_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PXI_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PXIE_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_MPCIE_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_CPCI_COP, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCI_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCIE_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCI104_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PCIE104_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PXI_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_PXIE_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_MPCIE_OPT, },
    { .vendor = ACTEL_VID, .device = ACTEL_DID, .subvendor = PCI_ANY_ID, .subdevice = FW_SSDID_CPCI_OPT, },
    { 0, },
};
#endif
MODULE_DEVICE_TABLE(pci, fw_idtable);

static struct pci_driver fw_pcidrv = {
    .name = DRV_NAME,
    .id_table = fw_idtable,
    .probe = fw_probe,
    .remove = fw_remove,
};

static int __init fw_init(void)
{
    int err = 0;

    printk(KERN_DEBUG DRV_NAME ": ==> Init\n");

    /* create device class - ls /sys/class */
    fw_class = class_create(THIS_MODULE, DRV_NAME);
    if (IS_ERR(fw_class)) {
        err = PTR_ERR(fw_class);
        goto fail;
    }

    /* allocate device numbers - cat /proc/devices */
    if (alloc_chrdev_region(&fw_dev_first, 0, max_devices, DRV_NAME) < 0) {
        err = -EIO;
        goto fail_class;
    }

    /* get major */
    pci_major = MAJOR(fw_dev_first);

    /* registering driver */
    return pci_register_driver(&fw_pcidrv);

fail_class:
    class_destroy(fw_class);
fail:
    return err;
}

static void __exit fw_exit(void)
{
    printk(KERN_DEBUG DRV_NAME ": => Exit\n");
    /* unregister driver */
    pci_unregister_driver(&fw_pcidrv);
    /* destroy device class */
    class_destroy(fw_class);
    /* free devices numbers */
    unregister_chrdev_region(fw_dev_first, max_devices);
}

module_init(fw_init);
module_exit(fw_exit);

MODULE_AUTHOR("Exoligent");
MODULE_DESCRIPTION("FIP Watcher Linux Driver");
MODULE_LICENSE("GPL v2");
MODULE_VERSION(FW_DRV_VERSION);