/*
 * Copyright (C) 2019-2023 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/param.h>
#include <linux/jiffies.h>
#include <linux/uaccess.h>
#include "fw-pci.h"
#include "../../include/fw-board.h"

#define DELAY 1

void wait_for_us(uint32_t delay_us)
{
    unsigned long delay_jiffies;
    bool wait;

    delay_jiffies = jiffies + usecs_to_jiffies(delay_us);
    wait = true;
    do {
        if (time_after(jiffies, delay_jiffies))
            wait = false;
    } while (wait);
}

int wait_for_reg32(uint32_t *reg, uint32_t flags, uint32_t timeout_us,
        uint8_t trigger_level)
{
    ktime_t end = ktime_add_us(ktime_get(), timeout_us);
    uint32_t cmp_flags = (trigger_level) ? flags : 0;
    /* polling method */
    for (;;)
    {
        if ((ioread32(reg) & flags) == cmp_flags)
            break;

        if (ktime_after(ktime_get() /* now */, end))
            return -EBUSY;
    }
    return 0;
}

void eeprom_enable(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);
    reg_data |= PLX_CTRL_EE_CS_HI;
    iowrite32(reg_data, reg_addr);
    reg_data = ioread32(reg_addr);
}

void eeprom_disable(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);
    reg_data &= ~PLX_CTRL_EE_CS_HI;
    iowrite32(reg_data, reg_addr);
}

void eeprom_do_clock_cycle(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);
    reg_data &= ~PLX_CTRL_EE_CLK_HI;
    iowrite32(reg_data, reg_addr);
    reg_data = ioread32(reg_addr);
    reg_data |= PLX_CTRL_EE_CLK_HI;
    iowrite32(reg_data, reg_addr);
    reg_data = ioread32(reg_addr);
    reg_data &= ~PLX_CTRL_EE_CLK_HI;
    iowrite32(reg_data, reg_addr);
}

uint8_t eeprom_read(struct fw_dev *fw, uint8_t *buffer, uint8_t start_address, uint8_t length)
{
    // start_address : address of the 1st 16-bit word to read
    // length : amount of 16-bit words to read
    uint32_t data;
    uint8_t mask, counter, i, j, k;
    uint8_t jo;

    counter = 0;
    eeprom_enable(fw);

    eeprom_write_data_bit_high(fw); /* Q = 1 */
    eeprom_write_data_bit_high(fw); /* Q = 1 */
    eeprom_write_data_bit_low(fw);  /* Q = 0 */

    mask = 0x80; /* mask = byte to send (A7 in first) */
    for (k = 0; k < 8; k++) {
        data = start_address & mask;
        (data == mask) ? eeprom_write_data_bit_high(fw) : /* Q = 1 */
                         eeprom_write_data_bit_low(fw);   /* Q = 0 */
        mask /= 2;
    }

    /* dummy bit test (should be 0) */
    jo = 0;
    while ((eeprom_read_data_bit_no_clock(fw) != 0) && (jo < 10))
        jo++;

    /* reading loop */
    for (i = 0; i < (2 * length); i++) {
        mask = 0x80;
        for (j = 8; j >= 1; j--) {
            data = eeprom_read_data_bit(fw);
            if (data == 0x01) /* bit read is 1 */
                *(buffer + i) |= mask;
            mask /= 2;
        }
        counter++;
    }
    eeprom_disable(fw);

    return counter;
}

void eeprom_write_data_bit_high(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);
    reg_data |= PLX_CTRL_EE_W_HI;
    iowrite32(reg_data, reg_addr);
    eeprom_do_clock_cycle(fw);
}

void eeprom_write_data_bit_low(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);
    reg_data &= ~PLX_CTRL_EE_W_HI;
    iowrite32(reg_data, reg_addr);
    eeprom_do_clock_cycle(fw);
}

uint8_t eeprom_read_data_bit(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data;
    eeprom_do_clock_cycle(fw);
    reg_data = ioread32(reg_addr);
    return ((reg_data & PLX_CTRL_EE_R_BIT) ? 0x01 : 0x00);
}

uint8_t eeprom_read_data_bit_no_clock(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t reg_data = ioread32(reg_addr);
    return ((reg_data & PLX_CTRL_EE_R_BIT) ? 0x01 : 0x00);
}

void init_interrupt(struct fw_dev *fw)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_INTCSR_REG);
    uint32_t reg_data = ioread32(reg_addr);

    /* enable LINTi1 (IRQn) and disable LINTi2 */
    reg_data |= PLX_INTCSR_LINT1_ENA;
    reg_data &= ~PLX_INTCSR_LINT2_ENA;

    /* LINTi1 (IRQn) - polarity 0: active low */
    reg_data &= ~PLX_INTCSR_LINT1_POL1;

    /* LINTi1 (IRQn) - level triggerable interrupt */
    reg_data &= ~PLX_INTCSR_LINT1_EDGE_ENA;

    /* local edge triggerable interrupt clear*/
    reg_data |= PLX_INTCSR_LINT1_EDGE_CLR;

    /* PCI interrupt enable */
    reg_data |= PLX_INTCSR_IT_ENA;

    iowrite32(reg_data, reg_addr);
}

int fw_irq_global_enable(struct fw_dev *fw, int enable)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_INTCSR_REG);
    uint32_t reg_val = ioread32(reg_addr);
    uint32_t reg_cmd = PLX_INTCSR_IT_ENA;

    if (enable)
        reg_val |= reg_cmd;
    else
        reg_val &= ~reg_cmd;

    iowrite32(reg_val, reg_addr);
    return 0;
}

int fw_irq_global_clear(struct fw_dev *fw, int clear)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_INTCSR_REG);
    uint32_t reg_val = ioread32(reg_addr);
    uint32_t reg_cmd = PLX_INTCSR_LINT1_EDGE_CLR;

    if (clear)
        reg_val |= reg_cmd;
    else
        reg_val &= ~reg_cmd;

    iowrite32(reg_val, reg_addr);
    return 0;
}

int fw_board_init(struct fw_dev *fw)
{
    uint32_t *res_addr = (uint32_t *) (fw->remap[BAR_BOARD] + RES_REG);
    uint32_t res_data = ioread32(res_addr);
    uint32_t *ctl_addr = (uint32_t *) (fw->remap[BAR_BOARD] + CTRL_REG);
    uint32_t ctl_data = ioread32(ctl_addr);
    uint8_t ctl_sta =  (uint8_t) ((ctl_data >> CTRL_STA_SHIFT) & CTRL_STA_MASK);
    uint8_t magic_num = (uint8_t) ((res_data >> RES_MAG_SHIFT) & RES_MAG_MASK);
    uint32_t bitrate = (uint32_t) ctl_sta & CTRL_STA_DATARATE;
    uint32_t sn = 0;

    if (magic_num != RES_MAG_NUM)
        goto public_part;

    /* private infos */
    /* => dma capability ? */
    /** TODO:
     * Resource registry to update inside the gateware
     * DMA is supported by fipwatcher board although the resource register
     * indicates that it does not.
     */
    //fw->has_dma_cap = (res_data & RES_DMA) ? true : false;
    /* dma is forced here */
    fw->has_dma_cap = true;

    if (fw->has_dma_cap) {

        /* dma mode 32-bit mode */
        if (dma_set_mask(&fw->pdev->dev, DMA_BIT_MASK(32)))
            fw->has_dma_cap = false;
        else {
            /* allocate a dma buffer (1MB) */
            fw->dma_buffer = kcalloc(1, DMA_BUFFER_BSZ, GFP_DMA);
            if (!fw->dma_buffer) {
                fw->has_dma_cap = false;
                goto public_part;
            }
        }
    }

public_part:
    // Get product number - read eeprom at 0x5e address (4 bytes)
    eeprom_read(fw, (uint8_t*) &sn, 0x5e, sizeof(sn)/2);

    /* public infos */
    fw->infos.index = fw->minor + 1;
    fw->infos.sn = sn;
    fw->infos.vid = fw->pdev->vendor;
    fw->infos.did = fw->pdev->device;
    fw->infos.ssvid = fw->pdev->subsystem_vendor;
    fw->infos.ssdid = fw->pdev->subsystem_device;
    fw->infos.irq_number = (int) fw->pdev->irq;
    fw->infos.drv_version = FW_DRV_MAJOR << 16;
    fw->infos.drv_version += FW_DRV_MINOR << 8;
    fw->infos.drv_version += FW_DRV_PATCH;
    fw->infos.drv_date = FW_DRV_BUILD_YEAR << 16;
    fw->infos.drv_date += FW_DRV_BUILD_MONTH << 8;
    fw->infos.drv_date += FW_DRV_BUILD_DAY;
    fw->infos.fip_bitrate = bitrate;
    return 0;
}

int fw_reset(struct fw_dev *fw, void __user *uarg)
{
    uint32_t *reg_addr = (uint32_t *) (fw->remap[BAR_BOARD] + PLX_CTRL_REG);
    uint32_t *ctl_addr = (uint32_t *) (fw->remap[BAR_BOARD] + CTRL_REG);
    uint32_t *sram_base_addr = (uint32_t *) (fw->remap[BAR_FW] + FW_MEM_OFFSET);
    uint32_t i, reg_data;

    /* disable irq */
    fw_irq_global_enable(fw, 0);

    /* reset global */
    reg_data = ioread32(ctl_addr);
    reg_data |= CTRL_CTL_RST_GLOBAL;
    iowrite32(reg_data, ctl_addr);
    wait_for_us(1000);   // 1ms waiting
    reg_data &= ~CTRL_CTL_RST_GLOBAL;
    iowrite32(reg_data, ctl_addr);

    reg_data = ioread32(reg_addr);
    reg_data |= PLX_CTRL_LOCAL_RST_ON;
    iowrite32(reg_data, reg_addr);
    wait_for_us(1000);   // 1ms waiting
    reg_data &= ~PLX_CTRL_LOCAL_RST_ON;
    iowrite32(reg_data, reg_addr);
    wait_for_us(100000); // 100ms waiting

    /* clean all sram memory + local buffer */
    for (i = 0; i < FW_MEM_WSZ; ++i) {
        iowrite32(0, sram_base_addr + i);
#if !defined(FW_VIA_MMAP)
        *(((uint32_t*) fw->irq.data) + i) = 0;
#endif
    }
    memset(fw->dma_buffer, 0, fw->dma_bsz);

    /* reset interruptions counters relative to the specified fipwatcher */
    fw->irq.src = 0;
    fw->irq.cnt = 0;
    fw->it_flag = 0;
    return 0;
}