/*
 * Copyright (C) 2019-2022 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#ifndef __FW_BOARD_H__
#define __FW_BOARD_H__

#if defined (__linux__)
#   include <asm/types.h>
#   include <linux/types.h>
#elif defined (_WIN32) || defined (_WIN64)
#   ifdef FW_WIN_DRV
#       include <ntddk.h>
#       define uint8_t UCHAR
#       define uint16_t USHORT
#       define uint32_t ULONG
#       define uint64_t ULONGLONG
#   else
#       include <windows.h>
#       include <winioctl.h>
#       include <stdint.h>
#   endif
#endif

/**
 * Note: FW_VIA_MMAP:
 * If this macro is set, the user apps has to implement mmap on /dev/mem
 * for fipwatcher device's BAR access. (else ioctl is used)
 */
//#define FW_VIA_MMAP

#define MAX_FW_BOARD 16
#define BAR_MAX 6
#define BAR_BOARD 0
#define BAR_FW 1
#define MAX_FW_CORE 4 // Max fipwatcher core inside one board
#define FW_MEM_OFFSET 0x8000u

#define FW_MEM_WSZ 0x2000u // Area size for one fipwatcher core
                           // 32768 bytes (32kB) => 8192 32-bits word
#define FW_MEM_BSZ ((FW_MEM_WSZ) * 4)

//**** OFFSETS (BAR0)
#define PLX_INTCSR_REG 0x4Cu
#define PLX_CTRL_REG 0x50u
#define PLX_GPIOC_REG 0x54u
// => Relative to Exoligent
/* resource register */
#define RES_REG 0x58u
/* fw control register */
#define CTRL_REG 0x5Cu
/* trigger & delay register */
#define TRIG_REG 0x60u
/* sensor register */
#define SNSR_REG 0x64u
/* epoch register */
#define EPOCH_LO_REG 0x68u
#define EPOCH_HI_REG 0x6Cu
/* dma registers */
#define DMA_AREA_PCI_OFFSET 0x400u
#define DMA_WSZ 0x10u
#define DMA_BSZ ((DMA_WSZ) * 4)
#define DMA_BASE_ADDR(x) ((x) * (DMA_BSZ))
#define DMA_SRC_PARAM_REG(x) (DMA_BASE_ADDR(x))
#define DMA_DST_PARAM_REG(x) ((DMA_BASE_ADDR(x)) + (0x04u))
#define DMA_SRC_ADDR_LO_REG(x) ((DMA_BASE_ADDR(x)) + (0x08u))
#define DMA_SRC_ADDR_HI_REG(x) ((DMA_BASE_ADDR(x)) + (0x0Cu))
#define DMA_DST_ADDR_LO_REG(x) ((DMA_BASE_ADDR(x)) + (0x10u))
#define DMA_DST_ADDR_HI_REG(x) ((DMA_BASE_ADDR(x)) + (0x14u))
#define DMA_LEN_REG(x) ((DMA_BASE_ADDR(x)) + (0x18u))
#define DMA_CTRL_REG(x) ((DMA_BASE_ADDR(x)) + (0x1Cu))
#define DMA_STA_REG(x) ((DMA_BASE_ADDR(x)) + (0x20u))
#define DMA_PRC_LEN_REG(x) ((DMA_BASE_ADDR(x)) + (0x24u))
#define DMA_SHR_ACC_REG(x) ((DMA_BASE_ADDR(x)) + (0x28u))
#define DMA_OP_R 0 /* read dma operation */
#define DMA_OP_W 1 /* write dma operation */
#define DMA_TMO_US 200000 /* dma timeout in us */
#define DMA_BUFFER_BSZ (1u << 20) /* 1MB */
/* msi registers */
#define MSI_ADDR_LO_REG 0x500u
#define MSI_ADDR_HI_REG 0x504u

/**
 * PLX_INTCSR_REG register: MASK & SHIFT
 */
enum fw_reg_plx_intcsr {
    PLX_INTCSR_LINT1_ENA = (1u << 0),
    PLX_INTCSR_LINT1_POL1 = (1u << 1),
    PLX_INTCSR_LINT1_ACT = (1u << 2),
    PLX_INTCSR_LINT2_ENA = (1u << 3),
    PLX_INTCSR_LINT2_POL1 = (1u << 4),
    PLX_INTCSR_LINT2_ACT = (1u << 5),
    PLX_INTCSR_IT_ENA = (1u << 6),
    PLX_INTCSR_SOFT_ENA = (1u << 7),
    PLX_INTCSR_LINT1_EDGE_ENA = (1u << 8),
    PLX_INTCSR_LINT2_EDGE_ENA = (1u << 9),
    PLX_INTCSR_LINT1_EDGE_CLR = (1u << 10),
    PLX_INTCSR_LINT2_EDGE_CLR = (1u << 11),
};

/**
 * PLX_CTRL_REG register: MASK & SHIFT
 */
enum fw_reg_plx_ctrl {
    PLX_CTRL_EE_CLK_HI = (1u << 24),
    PLX_CTRL_EE_CS_HI = (1u << 25),
    PLX_CTRL_EE_W_HI = (1u << 26),
    PLX_CTRL_EE_R_BIT = (1u << 27),
    PLX_CTRL_EE_R_PRES = (1u << 28),
    PLX_CTRL_EE_RELOAD_HI = (1u << 29),
    PLX_CTRL_LOCAL_RST_ON = (1u << 30),
};

/**
 * RES_REG register: MASK & SHIFT
 */
/* resource count */
#define RES_CNT_MASK    0x07u
#define RES_CNT_SHIFT   0u
/* dma capacity */
#define RES_DMA_MASK    0x01u
#define RES_DMA_SHIFT   19u
/* resource type (hardware) */
#define RES_TYPE_MASK   0x03u
#define RES_TYPE_SHIFT  14u
/* protocol version */
#define RES_PRTCL_MASK  0x0Fu
#define RES_PRTCL_SHIFT 16u
/* magic number */
#define RES_MAG_NUM     0x57
#define RES_MAG_MASK    0xFFu
#define RES_MAG_SHIFT   24u

enum fw_reg_res {
    RES_CNT = ((RES_CNT_MASK) << (RES_CNT_SHIFT)),
    RES_DMA = ((RES_DMA_MASK) << (RES_DMA_SHIFT)),
    RES_TYPE = ((RES_TYPE_MASK) << (RES_TYPE_SHIFT)),
    RES_PRTCL = ((RES_PRTCL_MASK) << (RES_PRTCL_SHIFT)),
    RES_MAG = ((RES_MAG_MASK) << (RES_MAG_SHIFT)),
};

/**
 * CTRL_REG register: MASK & SHIFT
 */
/* control */
#define CTRL_CTL_MASK   0xFFu
#define CTRL_CTL_SHIFT  0u
/* status */
#define CTRL_STA_MASK   0xFFu
#define CTRL_STA_SHIFT  8u
/* timeout/silence */
#define CTRL_TMO_MASK   0xFFu
#define CTRL_TMO_SHIFT  16u
/* trigger */
#define CTRL_TRIG_MASK  0xFFu
#define CTRL_TRIG_SHIFT 24u

enum fw_reg_ctrl {
    CTRL_CTL = ((CTRL_CTL_MASK) << (CTRL_CTL_SHIFT)),
    CTRL_STA = ((CTRL_STA_MASK) << (CTRL_STA_SHIFT)),
    CTRL_TMO = ((CTRL_TMO_MASK) << (CTRL_TMO_SHIFT)),
    CTRL_TRIG = ((CTRL_TRIG_MASK) << (CTRL_TRIG_SHIFT)),
};
// control: control
enum fw_reg_ctrl_ctl {
    CTRL_CTL_SEL = (0x03 << 0),
    CTRL_CTL_FIP_ENA = (1 << 2),
    CTRL_CTL_CLR_INT = (1 << 3),
    CTRL_CTL_SET_EPOCH = (1 << 4),
    CTRL_CTL_AUTO_RATE = (1 << 5),
    CTRL_CTL_RST_FIP = (1 << 6),
    CTRL_CTL_RST_GLOBAL = (1 << 7),
};
// control: status
enum fw_reg_ctrl_sta {
    CTRL_STA_DATARATE = (0x07 << 0),
    CTRL_STA_HW = (1 << 3),
    CTRL_STA_CARD = (1 << 4),
    CTRL_STA_OVF = (1 << 5),
    CTRL_STA_PID = (0x03 << 6),
};

/**
 * TRIG_REG register: MASK & SHIFT
 */
/* trigger delay */
#define TRIG_DELAY_MASK     0xFFFFu
#define TRIG_DELAY_SHIFT    0u
/* trigger id */
#define TRIG_ID_MASK        0xFFFFu
#define TRIG_ID_SHIFT       16u

enum fw_reg_trig {
    TRIG_DELAY = ((TRIG_DELAY_MASK) << (TRIG_DELAY_SHIFT)),
    TRIG_ID = ((TRIG_ID_MASK) << (TRIG_ID_SHIFT)),
};

/**
 * SNSR_REG register: MASK & SHIFT
 */
/* sensor value */
#define SNSR_DATA_MASK      0xFFFFu
#define SNSR_DATA_SHIFT     0u
/* sensor r0 */
#define SNSR_R0_MASK        1u
#define SNSR_R0_SHIFT       21u
/* sensor r1 */
#define SNSR_R1_MASK        1u
#define SNSR_R1_SHIFT       22u

enum fw_reg_snsr {
    SNSR_DATA = ((SNSR_DATA_MASK) << (SNSR_DATA_SHIFT)),
    SNSR_R0 = ((SNSR_R0_MASK) << (SNSR_R0_SHIFT)),
    SNSR_R1 = ((SNSR_R1_MASK) << (SNSR_R1_SHIFT)),
};

/**
 * DMA_SRC_PARAM_REG register: MASK & SHIFT
 */
/* dma read operation (vs write) */
#define DMA_SRC_PARAM_OP_R_MASK     0x01u
#define DMA_SRC_PARAM_OP_R_SHIFT    0u

enum fw_reg_dma_src_param {
    DMA_SRC_PARAM_OP_R = ((DMA_SRC_PARAM_OP_R_MASK) << (DMA_SRC_PARAM_OP_R_SHIFT)),
};

/**
 * DMA_CTRL_REG register: MASK & SHIFT
 */
/* start/abort dma operation */
#define DMA_CTRL_OP_STRT_ABRT_MASK  0x01u
#define DMA_CTRL_OP_STRT_ABRT_SHIFT 0u

enum fw_reg_dma_ctrl {
    DMA_CTRL_OP_STRT_ABRT = ((DMA_CTRL_OP_STRT_ABRT_MASK) << (DMA_CTRL_OP_STRT_ABRT_SHIFT)),
};

/**
 * DMA_STA_REG register: MASK & SHIFT
 */
/* dma completion ok */
#define DMA_STA_COMPL_OK_MASK   0x01u
#define DMA_STA_COMPL_OK_SHIFT  0u
/* dma completion with error */
#define DMA_STA_COMPL_ERR_MASK  0x01u
#define DMA_STA_COMPL_ERR_SHIFT 3u
/* dma watchdog error */
#define DMA_STA_ERR_WDT_MASK    0x01u
#define DMA_STA_ERR_WDT_SHIFT   5u
/* dma stopped by user */
#define DMA_STA_ERR_USR_MASK    0x01u
#define DMA_STA_ERR_USR_SHIFT   6u
/* dma buffer error */
#define DMA_STA_ERR_END_MASK    0x01u
#define DMA_STA_ERR_END_SHIFT   7u

enum fw_reg_dma_sta {
    DMA_STA_COMPL_OK = ((DMA_STA_COMPL_OK_MASK) << (DMA_STA_COMPL_OK_SHIFT)),
    DMA_STA_COMPL_ERR = ((DMA_STA_COMPL_ERR_MASK) << (DMA_STA_COMPL_ERR_SHIFT)),
    DMA_STA_ERR_WDT = ((DMA_STA_ERR_WDT_MASK) << (DMA_STA_ERR_WDT_SHIFT)),
    DMA_STA_ERR_USR = ((DMA_STA_ERR_USR_MASK) << (DMA_STA_ERR_USR_SHIFT)),
    DMA_STA_ERR_END = ((DMA_STA_ERR_END_MASK) << (DMA_STA_ERR_END_SHIFT)),
};

/* Subsystem Device ID (SSDID) */
enum fw_ssdid {
    /* copper physical layer */
    FW_SSDID_PCI_COP = 0x5202,
    FW_SSDID_PCIE_COP = 0x5302,
    FW_SSDID_PCI104_COP = 0x5402,
    FW_SSDID_PCIE104_COP = 0x5502,
    FW_SSDID_PXI_COP = 0x5602,
    FW_SSDID_PXIE_COP = 0x5702,
    FW_SSDID_MPCIE_COP = 0x5802,
    FW_SSDID_CPCI_COP = 0x5902,
    /* optical physical layer */
    FW_SSDID_PCI_OPT = 0x7202,
    FW_SSDID_PCIE_OPT = 0x7302,
    FW_SSDID_PCI104_OPT = 0x7402,
    FW_SSDID_PCIE104_OPT = 0x7502,
    FW_SSDID_PXI_OPT = 0x7602,
    FW_SSDID_PXIE_OPT = 0x7702,
    FW_SSDID_MPCIE_OPT = 0x7802,
    FW_SSDID_CPCI_OPT = 0x7902,
};

enum fw_board_format {
    _FW_BOARD_MIN = 2,
    FW_BOARD_PCI = _FW_BOARD_MIN,
    FW_BOARD_PCIE = 3,
    FW_BOARD_PCI104 = 4,
    FW_BOARD_PCIE104 = 5,
    FW_BOARD_PXI = 6,
    FW_BOARD_PXIE = 7,
    FW_BOARD_MPCIE = 8,
    FW_BOARD_CPCI = 9,
    _FW_BOARD_MAX,
};

#pragma pack(1) // 1 byte alignment

#if defined (_WIN32) || defined (_WIN64)
#elif defined (__linux__)
struct arg_io {
    uint64_t i_ptr;
    uint64_t o_ptr;
    uint32_t i_bsz;
    uint32_t o_bsz;
    uint32_t status;
};
#endif

enum fw_dev_irq_flags {
    FW_IRQ_LINT1 = (1u << 30),
    FW_IRQ_LINT2 = (1u << 31),
};

struct fw_dev_irq {
    uint32_t src;
    uint64_t cnt;
#if !defined(FW_VIA_MMAP)
    uint32_t data[FW_MEM_WSZ]; /* fipwatcher buffer received */
#endif
};

struct fw_dev_infos {
    uint32_t    index;
    uint32_t    sn; /**< serial number */
    /* device general info */
    uint16_t    vid;
    uint16_t    did;
    uint16_t    ssvid;
    uint16_t    ssdid;
    /* physical BAR addresses */
    int         bar_cnt; /**< number of BAR for this device */
    uint64_t    bar_bsz[BAR_MAX]; /**< size of the BAR area */
    uint64_t    bar_base[BAR_MAX]; /**< memory BAR (start phy address) */
    /* irq infos*/
    int         irq_number;
    /* driver infos */
    uint32_t    drv_version; /**< driver version */
    uint32_t    drv_date; /**< driver build date */
    /* fip infos */
    int         fip_bitrate; /**< see enum fipcap_bitrate */
};

struct fw_act_set_cmd {
    uint32_t op; /* 0: sniffer off, 1: sniffer on */
};

struct fw_epoch_set_cmd {
    uint64_t epoch; /* new epoch to apply */
};

#pragma pack()

//***************************************************************************************
//                              IOCTLs Definitions
//***************************************************************************************
enum fw_ioctl_cmd {
#if defined (_WIN32) || defined (_WIN64)
    FW_INFO_GET = 0x801,
#else
    FW_INFO_GET,
#endif
    FW_ACT_SET,
    FW_EPOCH_SET,
    FW_RESET,
    FW_IRQ_WAIT_FOR,
    FW_IRQ_RESUME,
};

#if defined (_WIN32) || defined (_WIN64)
#    define C_FIP_TYPE 0x8C0F
#    define FW_IOCTL_INFO_GET CTL_CODE(C_FIP_TYPE, FW_INFO_GET, METHOD_BUFFERED, FILE_READ_ACCESS)
#    define FW_IOCTL_ACT_SET CTL_CODE(C_FIP_TYPE, FW_ACT_SET, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#    define FW_IOCTL_EPOCH_SET CTL_CODE(C_FIP_TYPE, FW_EPOCH_SET, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#    define FW_IOCTL_RESET CTL_CODE(C_FIP_TYPE, FW_RESET, METHOD_BUFFERED, FILE_ANY_ACCESS)
#    define FW_IOCTL_IRQ_WAIT_FOR CTL_CODE(C_FIP_TYPE, FW_IRQ_WAIT_FOR, METHOD_BUFFERED, FILE_READ_ACCESS)
#    define FW_IOCTL_IRQ_RESUME CTL_CODE(C_FIP_TYPE, FW_IRQ_RESUME, METHOD_BUFFERED, FILE_ANY_ACCESS)
#elif defined (__linux__)
#    define FW_IOCTL_MAGIC 0xFF
#    define FW_IOCTL_INFO_GET _IOR(FW_IOCTL_MAGIC, FW_INFO_GET, struct fw_dev_infos)
#    define FW_IOCTL_ACT_SET _IOW(FW_IOCTL_MAGIC, FW_ACT_SET, struct fw_act_set_cmd)
#    define FW_IOCTL_EPOCH_SET _IOW(FW_IOCTL_MAGIC, FW_EPOCH_SET, struct fw_epoch_set_cmd)
#    define FW_IOCTL_RESET _IO(FW_IOCTL_MAGIC, FW_RESET)
#    define FW_IOCTL_IRQ_WAIT_FOR _IOWR(FW_IOCTL_MAGIC, FW_IRQ_WAIT_FOR, struct arg_io)
#    define FW_IOCTL_IRQ_RESUME _IO(FW_IOCTL_MAGIC, FW_IRQ_RESUME)
#endif

#endif