/*
 * Copyright (C) 2023 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#ifndef __COMMON_H__
#define __COMMON_H__

#if defined(__QNX__)
#   include <sys/syspage.h> // SYSPAGE_ENTRY
#endif

#if defined (_WIN32) || defined (_WIN64)
#   include <windows.h>
/* Some MinGW/CYGWIN distributions don't define the macro: */
#   ifndef ENABLE_VIRTUAL_TERMINAL_PROCESSING
#       define ENABLE_VIRTUAL_TERMINAL_PROCESSING 0x0004
#   endif
#endif

#include <sys/time.h>   // gettimeofday()
#include <time.h>       // localtime()

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>     // memset()
#include <stdint.h>     // uint8_t, ...
#include <inttypes.h>   // PRIx64
#include <stdbool.h>    // bool
#include <pthread.h>    // pthread()
#include <unistd.h>     // usleep()
#include <stdarg.h>     // args

#include "libfipcap.h"

#define APP_VERSION "1.0.0"

enum tst_verbose_level {
    VERBOSE_LEVEL_NONE = 0,
    VERBOSE_LEVEL_LOW,
    VERBOSE_LEVEL_MEDIUM,
    VERBOSE_LEVEL_HIGH,
};

#pragma pack(1) // 1 byte alignment

struct tst_measure {
    /* synchro callback measure */
    uint64_t nb_sync_cb;
    uint64_t cumul_sync_cb_duration;
    uint64_t min_sync_cb_duration;
    uint64_t max_sync_cb_duration;
    uint64_t min_sync_cb_last;
    uint64_t max_sync_cb_last;
};

struct tst_session {
    /* pci device handle */
    struct fipcap_dev *dev;
    /* fip capture session */
    struct fipcap_sess *cap_s;
    /* performance measurement */
    struct tst_measure stats;
    /* verbosity */
    int verbose_level; /* enum tst_verbose_level */
};
#pragma pack()

extern int running;
extern bool dbg_coloriz;

/****************************************************************************************
 * FIP CAPTURE SESSION FUNCTIONS
 ***************************************************************************************/
extern int tst_session_start(struct tst_session *sess);
extern int tst_session_stop(struct tst_session *sess);

/****************************************************************************************
 * FIP BOARDS FUNCTIONS
 ***************************************************************************************/
extern int dev_open(int dev_idx, struct tst_session *sess);
extern int dev_close(struct tst_session *sess);
extern void dev_show_all();

/****************************************************************************************
 * STATISTICS
 ***************************************************************************************/
extern void update_min_max(uint64_t current_value, uint64_t *min_value,
            uint64_t *max_value);
extern int _usleep(uint32_t usec);
#if defined(__QNX__) || defined(_WIN32) || defined(_WIN64)
extern int _gettime(uint64_t *when);
extern void update_stats(uint64_t *start, uint64_t *stop, uint64_t *cp,
    uint64_t *cumul, uint64_t *min, uint64_t *max);
#else
extern int _gettime(struct timeval *when);
extern void update_stats(struct timeval *start, struct timeval *stop,
    uint64_t *cp, uint64_t *cumul, uint64_t *min, uint64_t *max);
#endif
extern char *epoch_to_str(uint64_t epoch_us);
extern char *duration_to_str(uint64_t duration_us);

/****************************************************************************************
 * TIME/DATE FUNCTIONS
 ***************************************************************************************/
#define DBG_PRINT_INFO 1
#define DBG_PRINT_WARN 1
#define DBG_PRINT_ERR  1
#define DBG_PRINT_EVT  1
#define DBG_DUMP_MEM   0
#define DGB_COLORIZ    1

typedef enum
{
    cap = 0,    // capture (sniffer)
    app,        // application (general)
    no_src      // unknown
} msg_src;

typedef enum
{
    info = 0,   // information
    warn,       // warning
    err,        // error
    evt,        // event (callback)
    no_type     // unknown
} msg_type;

extern char *get_date();
extern struct timespec timespec_diff(struct timespec start, struct timespec end);
extern struct timespec timespec_add_ns(struct timespec ref, long ns);
extern bool b_then_a(struct timespec a, struct timespec b);

#if defined(__QNX__)
extern uint64_t clock_diff_ns(uint64_t start_clk_cycl, uint64_t end_clk_cycl);
#endif

#if defined(__QNX__) || defined(_WIN32) || defined(_WIN64)
extern unsigned long get_elapsed_us(uint64_t end, uint64_t start);
#else
extern unsigned long get_elapsed_us(struct timeval end, struct timeval start);
#endif

/****************************************************************************************
 * CUSTOMIZED PRINT FUNCTION
 ***************************************************************************************/
extern int _print(msg_src src, msg_type type, const char *fmt, ...);

#if defined (_WIN32) || defined (_WIN64)
extern void enable_console_colorization(); /* enable console colorization for win */
#endif

#endif