/*
 * Copyright (C) 2023 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 *
 * Description:
 * Capture all frames from FIP network (using libfipcap library)
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>     // memset()
#include <stdint.h>     // uint8_t, ...
#include <fcntl.h>      // open()
#include <unistd.h>     // read()
#include <libgen.h>     // basename
#include <inttypes.h>   // PRIx64+
#include <signal.h>     // signal
#include <time.h>       // nanosleep, timespec
#if defined (_WIN32) || defined (_WIN64)
#include <conio.h>      // kbhit()
#endif

#include "common.h"

char *app_name = NULL;
int running = 0;

static void help();
void handle_signal(int sig);

int main(int argc, char *argv[])
{
    int ret = 0;
    int dev_idx = 1;
    int value = 0;
    int verbose_level = VERBOSE_LEVEL_NONE;
    struct timespec rqtp = {0};
    struct timespec rmtp = {0};
    struct tst_session sess;

#if defined (_WIN32) || defined (_WIN64)
    SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
    enable_console_colorization();
#else
    uint8_t buff = 0;
#endif

    app_name = basename(argv[0]);

    /* Try to process all command line arguments */
    while ((value = getopt(argc, argv, "i:lhv")) != -1) {
        switch (value) {
            case 'i':
                sscanf(optarg, "%d", &dev_idx);
                break;
            case 'l':
                dev_show_all();
                return EXIT_SUCCESS;
            case 'h':
                help();
                return EXIT_SUCCESS;
            case 'v':
                verbose_level += \
                    (verbose_level > VERBOSE_LEVEL_MEDIUM) ? 0 : 1;
                break;
            case '?':
            default:
                help();
                return EXIT_FAILURE;
        }
    }

    signal(SIGINT, handle_signal);

    /* fipcap - initialization [mandatory] */
    if (fipcap_init()) {
        _print(app, err, "cannot init fipcap library: %s\n",
            fipcap_strerror(errno));
        return -1;
    }

    /* exoligent - fip device opening */
    memset(&sess, 0, sizeof(struct tst_session));
    if ((ret = dev_open(dev_idx, &sess)))
        goto fail1;

    sess.verbose_level = verbose_level;

    if (tst_session_start(&sess) < 0) {
        _print(app, err, "impossible to start fip capture.\n");
        goto fail2;
    }

    _print(app, info, "app started. press enter to close...\n");

#if defined(__linux__) || defined(__QNX__)
    /* set stdin non-blocking */
    fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
#endif

    rqtp.tv_sec = 0;
    rqtp.tv_nsec = 250000000; /* 250ms */

    /* main loop */
    running = 1;
    while (running) {

        /* wait for a pressed key */
#if defined (_WIN32) || defined (_WIN64)
        if (kbhit())
#else
        if (read(STDIN_FILENO, &buff, 1) == 1)
#endif
            running = 0;

        ret = nanosleep(&rqtp, &rmtp);
    }
    printf("\n");

    if (tst_session_stop(&sess))
        _print(app, err, "impossible to stop fip capture.\n");

fail2:
    /* exoligent - fip device closure */
    if (dev_close(&sess))
        _print(app, err, "cannot close correctly the device.\n");
fail1:
    fipcap_exit();
    _print(app, info, "app closed\n");
    return ret;
}

/**
 * \brief Callback function for handling signals.
 * \param sig identifier of signal
 */
void handle_signal(int sig)
{
    if (sig == SIGINT) {
        _print(app, info, "SIGINT signal received\n");
        running = 0;
        /* Reset signal handling to default behavior */
        signal(SIGINT, SIG_DFL);
    } else
        _print(app, warn, "signal received but not supported (sig=%d)\n", sig);
}

/**
 * \brief Print help for this application
 */
static void help()
{
    printf("\n");
    printf("Name: %s\n", app_name);
    printf("Version: %s\n\n", APP_VERSION);
    printf("Usage: %s [OPTION]...\n\n", app_name);
    printf("It captures all the frames of the FIP network.\n");
    printf("By default, if no option is added, the app opens the first PCI/PCIe device\n");
    printf("with index 1 [-i 1] and statistics verbosity.\n\n");
    printf("Options:\n");
    printf("    -i   device index [default=1]\n");
    printf("    -l   list the FIP boards present on the host machine\n");
    printf("    -v   be verbose (-vv or -vvv for higher verbosity)\n");
    printf("    -h   show version and this help and exit\n");
    printf("\n");
    printf("Examples: \n");
    printf("    /* statistics verbosity */\n");
    printf("    %s\n\n", app_name);
    printf("    /* raw frames verbosity */\n");
    printf("    %s -vv\n\n", app_name);
    printf("    /* detailed frames verbosity */\n");
    printf("    %s -vvv\n", app_name);
    printf("\n");
}