/*
 * Copyright (C) 2023 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#include "common.h"
#include "session.h"

int tst_session_start(struct tst_session *sess)
{
    const struct fipcap_version *lib_version = fipcap_version_get();
    struct fipcap_sess_cfg cap_s_cfg;

    memset(&cap_s_cfg, 0, sizeof(struct fipcap_sess_cfg));

    /* fipcap - versions */
    _print(app, info, "test: v%s - fipcap lib: v%d.%d.%d\n", APP_VERSION,
        lib_version->info.major, lib_version->info.minor, lib_version->info.patch);

    _print(app, info, "[fip] capture configuration\n");
    /* set capture config */
    /* hardware access */
    cap_s_cfg.dev = sess->dev; /* attach pci device to node configuration */

    /* app handlers */
    cap_s_cfg.fipcap_frm_handler = tst_fipcap_frm_handler;
    /* note: If the user has not set any verbosity on the frames,
     * we display the session report at regular intervals */
    cap_s_cfg.fipcap_diag_handler = \
        (sess->verbose_level == VERBOSE_LEVEL_NONE) ? \
            tst_fipcap_diag_handler : NULL;
    cap_s_cfg.fipcap_error_handler = tst_fipcap_error_handler;
    cap_s_cfg.fipcap_reset_handler = tst_fipcap_reset_handler;

    /**
     * User app context
     * note: Useful to retrieve the session context from a fipcap callback
     */
    cap_s_cfg.user_ctx = (void*) sess;

    _print(cap, info, "[fip] capture init\n");
    sess->cap_s = fipcap_sess_init(&cap_s_cfg);
    if (!sess->cap_s) {
       _print(cap, err, "capture initialization failed: %s\n",
            fipcap_strerror(errno));
        return -1;
    }

    print_cap_infos(sess->cap_s);

    _print(cap, info, "[fip] capture start\n");
    if (fipcap_sess_start(sess->cap_s)) {
        _print(cap, err, "capture starting failed: %s\n",
            fipcap_strerror(errno));
        goto fail;
    }

    return 0;

fail:
    fipcap_sess_exit(sess->cap_s);
    return -1;
}

int tst_session_stop(struct tst_session *sess)
{
    struct fipcap_sess_report r;

    memset(&r, 0, sizeof(struct fipcap_sess_report));

    /* stop the capture */
    if (fipcap_sess_stop(sess->cap_s)) {
        _print(cap, err, "capture stopping failed: %s\n",
            fipcap_strerror(errno));
    } else
        _print(cap, info, "[fip] capture stop ok\n");

    /* print capture report */
    if (fipcap_sess_report_get(sess->cap_s, &r)) {
        _print(cap, err, "capture report: %s\n",
            fipcap_strerror(errno));
    } else
        tst_fipcap_diag_handler(sess->cap_s, &r);

    /* print session stats */
    _print(cap, info, "** FRAME HANDLER CALLS     : TOTAL=%"PRIu64"\n",
        sess->stats.nb_sync_cb);

    if (sess->stats.nb_sync_cb) {
        uint64_t avg_duration = ((float) sess->stats.cumul_sync_cb_duration) \
                    / sess->stats.nb_sync_cb;
        _print(cap, info,
            "    occurrence period  (us): min=%"PRIu64", max=%"PRIu64"\n",
            sess->stats.min_sync_cb_last,
            sess->stats.max_sync_cb_last);
        _print(cap, info,
            "    treatment duration (us): min=%"PRIu64", avg=%"PRIu64", max=%"PRIu64"\n\n",
            sess->stats.min_sync_cb_duration, avg_duration,
            sess->stats.max_sync_cb_duration);
    }

    /* delete all resources attached */
    if (fipcap_sess_exit(sess->cap_s)) {
        _print(cap, err, "capture exiting failed: %s\n",
            fipcap_strerror(errno));
        return -1;
    }
    _print(cap, info, "[fip] capture exit ok\n");
    return 0;
}

void tst_fipcap_error_handler(struct fipcap_sess *cap_s,
        enum fipcap_error_code code)
{
    if (running)
        _print(cap, err,
            "error_handler: %s (err_code=%d)\n",
            fipcap_strerror(code),
            code);
    running = 0;
}

void tst_fipcap_reset_handler(struct fipcap_sess *cap_s)
{
    if (fipcap_device_reset(cap_s->infos->cfg.dev)) {
        _print(cap, err, "reset failed: %s\n", fipcap_strerror(errno));
        return;
    }
    _print(cap, evt, "reset component (fipwatcher)\n");
}

void tst_fipcap_frm_handler(struct fipcap_sess *cap_s,
        struct fipcap_frm *frm)
{
    struct tst_session *sess = (struct tst_session *) cap_s->infos->cfg.user_ctx;

    //unsigned long duration_us = 0;
    unsigned long last_us = 0;
    static int is_first = 1;
#if defined(__QNX__) || defined(_WIN32) || defined(_WIN64)
    static uint64_t start_treatment, stop_treatment, last_treatment = 0;
#else
    static struct timeval start_treatment, stop_treatment,
            last_treatment = { .tv_sec = 0, .tv_usec = 0, };
#endif
    int i;
    char *ctrl_str;

    /**
     * We avoid logging the warnings at app startup and shutdown:
     * - at startup: to give the remote node time to write its variables
     *   at least once.
     * - at shutdown: to avoid not significant warns.
     */
    //int log_warn = (!running || is_first) ? 0 : 1;

    _gettime(&start_treatment);

    if (sess->verbose_level == VERBOSE_LEVEL_NONE)
        goto shunt_print;

    switch(frm->ctrl) {
    case FIPCAP_FRM_CTRL_ID_DAT: ctrl_str = "ID_DAT"; break;
    case FIPCAP_FRM_CTRL_ID_MSG: ctrl_str = "ID_MSG"; break;
    case FIPCAP_FRM_CTRL_ID_RQ1: ctrl_str = "ID_RQ1"; break;
    case FIPCAP_FRM_CTRL_ID_RQ2: ctrl_str = "ID_RQ2"; break;
    case FIPCAP_FRM_CTRL_RP_DAT: ctrl_str = "RP_DAT"; break;
    case FIPCAP_FRM_CTRL_RP_DAT_MSG: ctrl_str = "RP_DAT_MSG"; break;
    case FIPCAP_FRM_CTRL_RP_DAT_RQ1: ctrl_str = "RP_DAT_RQ1"; break;
    case FIPCAP_FRM_CTRL_RP_DAT_RQ2: ctrl_str = "RP_DAT_RQ2"; break;
    case FIPCAP_FRM_CTRL_RP_DAT_RQ1_MSG: ctrl_str = "RP_DAT_RQ1_MSG"; break;
    case FIPCAP_FRM_CTRL_RP_DAT_RQ2_MSG: ctrl_str = "RP_DAT_RQ2_MSG"; break;
    case FIPCAP_FRM_CTRL_RP_RQ1: ctrl_str = "RP_RQ1"; break;
    case FIPCAP_FRM_CTRL_RP_RQ2: ctrl_str = "RP_RQ2"; break;
    case FIPCAP_FRM_CTRL_RP_MSG_NOACK: ctrl_str = "RP_MSG_NOACK"; break;
    case FIPCAP_FRM_CTRL_RP_MSG_ACK_EVEN: ctrl_str = "RP_MSG_ACK_EVEN"; break;
    case FIPCAP_FRM_CTRL_RP_MSG_ACK_ODD: ctrl_str = "RM_MSG_ACK_ODD"; break;
    case FIPCAP_FRM_CTRL_RP_ACK_EVEN_P: ctrl_str = "RP_ACK_EVEN_P"; break;
    case FIPCAP_FRM_CTRL_RP_ACK_ODD_P: ctrl_str = "RP_ACK_ODD_P"; break;
    case FIPCAP_FRM_CTRL_RP_ACK_EVEN_M: ctrl_str = "RP_ACK_EVEN_M"; break;
    case FIPCAP_FRM_CTRL_RP_ACK_ODD_M: ctrl_str = "RP_ACK_ODD_M"; break;
    case FIPCAP_FRM_CTRL_RP_FIN: ctrl_str = "RP_FIN"; break;
    default: ctrl_str = "UNKNOWN"; break;
    }

    if (sess->verbose_level > VERBOSE_LEVEL_NONE)
    {
        if (frm->error)
            _print(cap, err, "(%" PRIu64 ") %s(0x%04x) [%s]\n",
                frm->idx, ctrl_str, frm->id, frm_err_str[frm->error]);
        else
            _print(cap, evt, "(%" PRIu64 ") %s(0x%04x)\n",
                frm->idx, ctrl_str, frm->id);
    }

    if (sess->verbose_level == VERBOSE_LEVEL_MEDIUM)
    {
        /* raw frame print */
        printf("raw_data: (bsz=%d)\n", frm->raw_bsz);
        for (i = 0; i < frm->raw_bsz; ++i) {
            printf("%02x ", *(frm->raw + i));
        }
        printf("\n");
    }
    else if (sess->verbose_level == VERBOSE_LEVEL_HIGH)
    {
        printf("frame epoch: %s\n", epoch_to_str(frm->time_info.epoch));
        printf("frame duration: %.1f us\n",
            ((double) frm->time_info.frame_nstime) / 1000.);
        printf("inter-frame duration: %.1f us\n",
            ((double) frm->time_info.interframe_nstime) / 1000.);

        /* fip variable frame print (ie: RP_DAT_XXX frames) */
        if (frm->usrdata.type == FIPCAP_FRM_USRDATA_VAR) {
            printf("pdu: 0x%02x\n", frm->usrdata.var.pdu);
            printf("data: (bsz=%d)\n", frm->usrdata.var.payload_bsz);
            for (i = 0; i < frm->usrdata.var.payload_bsz; ++i)
                printf("%02x ", *(frm->usrdata.var.payload + i));
            printf("\n");
        }

        /* fip message frame print (ie: RP_MSG_XXX frames) */
        if (frm->usrdata.type == FIPCAP_FRM_USRDATA_MSG) {
            printf("src: addr=%08x (seg=%d lsap=0x%04x)\n",
                frm->usrdata.msg.hdr.src.addr,
                frm->usrdata.msg.hdr.src.seg, frm->usrdata.msg.hdr.src.lsap);
            printf("dst: addr=%08x (seg=%d lsap=0x%04x)\n",
                frm->usrdata.msg.hdr.dst.addr,
                frm->usrdata.msg.hdr.dst.seg, frm->usrdata.msg.hdr.dst.lsap);
            printf("data: (bsz=%d)\n", frm->usrdata.msg.payload_bsz);
            for (i = 0; i < frm->usrdata.msg.payload_bsz; ++i)
                printf("%02x ", *(frm->usrdata.msg.payload + i));
            printf("\n");
        }

        /* var_list frame print (ie: RP_RQ1, RP_RQ2 frames) */
        if (frm->usrdata.type == FIPCAP_FRM_USRDATA_VAR_LIST) {
            printf("ids_cnt: %d\n", frm->usrdata.var_list.ids_cnt);
            for (i = 0; i < frm->usrdata.var_list.ids_cnt; ++i) {
                uint16_t id = frm->usrdata.var_list.ids[i];
                /* bswap16 */
                id = ((id << 8) & 0xff00) | ((id >> 8) & 0x00ff);
                printf(" - %04x\n", id);
            }
        }

        printf("frame check sequence: 0x%04x\n", frm->fcs);
    }

shunt_print:

    _gettime(&stop_treatment);
    //duration_us = get_elapsed_us(stop_treatment, start_treatment);
    last_us = (is_first) ? 0 : get_elapsed_us(start_treatment, last_treatment);
    is_first = 0;
    last_treatment = start_treatment;

    /**
     * Update session stats
     */
    /* => min/max interval between two network events (synch. var) */
    if (last_us)
        update_min_max(last_us, &sess->stats.min_sync_cb_last,
            &sess->stats.max_sync_cb_last);
    /* => event treatment duration */
    update_stats(&start_treatment, &stop_treatment, &sess->stats.nb_sync_cb,
        &sess->stats.cumul_sync_cb_duration,
        &sess->stats.min_sync_cb_duration,
        &sess->stats.max_sync_cb_duration);

    //_print(cap, evt, "synchro callback [duration=%ldus - last=%ldus ago]\n",
    //   duration_us, last_us);
}

void tst_fipcap_diag_handler(struct fipcap_sess *sess,
        struct fipcap_sess_report *r)
{
    int i, j;

    if (r->status.state == FIPCAP_SESS_STATE_RUNNING) {
#if defined(_WIN32) || defined(_WIN64)
        system("cls");
#else
        if (system("clear"))
            return;
#endif
        _print(cap, info, "PRESS ENTER TO CLOSE ...\n\n");
    } else
        _print(cap, info, "** CAPTURE SESSION   :\n");

    _print(cap, info, "  SESSION INFOS      :\n");
    _print(cap, info, "    ** network information\n");
    _print(cap, info, "    frame_type       : %s\n",
        frm_type_str[r->status.frm_type]);
    _print(cap, info, "    frame_bitrate    : %s\n",
        frm_bitrate_str[r->status.frm_bitrate]);
    _print(cap, info, "    ** capturing time\n");
    _print(cap, info, "    start epoch      : %s\n",
        (r->status.start_epoch) ? epoch_to_str(r->status.start_epoch) : "-");
    _print(cap, info, "    stop epoch       : %s\n",
        (r->status.stop_epoch) ? epoch_to_str(r->status.stop_epoch) : "-");
    _print(cap, info, "    duration         : %s us\n",
        (r->status.duration_us) ? duration_to_str(r->status.duration_us) : "0");
    _print(cap, info, "    ** fip time\n");
    _print(cap, info, "    silence          : %d us\n", r->status.silence_ustime);
    _print(cap, info, "    turn_around      : %d us\n", r->status.turn_around_ustime);
    _print(cap, info, "    ** bus arbiter infos\n");
    _print(cap, info, "    fip address      : 0x%02x\n", r->ba.addr);
    _print(cap, info, "    mcycle index     : %d\n", r->ba.mcycle_idx);
    _print(cap, info, "    mcycle duration  : %d us\n", r->ba.mcycle.duration_ustime);
    _print(cap, info, "    %d window(s)      :\n", r->ba.mcycle.wind_cnt);

    for (i = 0; i < r->ba.mcycle.wind_cnt; ++i)
    {
        int wind_type = r->ba.mcycle.wind_list[i].type;
        _print(cap, info, "    - type           : %s\n",
            ba_wind_type_str[wind_type]);

        if (wind_type == FIPCAP_BA_WIND_PER_VAR ||
            wind_type == FIPCAP_BA_WIND_PER_MSG)
        {
            int req_cnt = r->ba.mcycle.wind_list[i].per.req_cnt;
            for (j = 0; j < req_cnt; ++j)
            {
                _print(cap, info, "      + %s(0x%04x)\n",
                    ba_id_type_str[r->ba.mcycle.wind_list[i].per.req_list[j].type],
                    r->ba.mcycle.wind_list[i].per.req_list[j].id);
            }
        }
        else if (wind_type == FIPCAP_BA_WIND_APER_VAR ||
                 wind_type == FIPCAP_BA_WIND_APER_MSG)
        {
            _print(cap, info, "      + end time     : %d us\n",
                r->ba.mcycle.wind_list[i].aper.end_ustime);
        }
    }

    _print(cap, info, "    ** nodes infos\n");

    for (i = 0, j = 0; i < _FIPCAP_NODE_CNT_MAX; ++i) {
        if (r->node[i].is_present)
            j++;
    }

    _print(cap, info, "    %d node(s)        :\n", j);

    for (i = 0; i < _FIPCAP_NODE_CNT_MAX; ++i) {
        struct fipcap_sess_node_info *n_i = &r->node[i];

        if (!n_i->is_present)
            continue;

        _print(cap, info, "    - node address   : 0x%02x\n", i);

        if (r->status.state == FIPCAP_SESS_STATE_RUNNING)
            goto shunt_node_details;

        /* identification infos */
        _print(cap, info, "      + manufacturer : %s\n",
            n_i->manufacturer_name);
        _print(cap, info, "      + model_name   : %s\n",
            n_i->model_name);
        _print(cap, info, "      + tag_name     : %s\n",
            n_i->tag_name);

        /* presence infos */
        _print(cap, info, "      + last present : %s\n",
            (n_i->last_pres_epoch) ? epoch_to_str(n_i->last_pres_epoch) : "-");
        _print(cap, info, "      + last absence : %s\n",
            (n_i->last_abs_epoch) ? epoch_to_str(n_i->last_abs_epoch) : "-");
        _print(cap, info, "      + absence count: %d\n",
            n_i->absence_cnt);

        if ((n_i->ba_status == FIPCAP_BA_STATUS_ACTIVE) ||
            (n_i->ba_status == FIPCAP_BA_STATUS_IDLE))
            _print(cap, info, "      + ba_prio      : %d\n",
                n_i->ba_prio);

shunt_node_details:
        _print(cap, info, "      + ba_status    : %s\n",
            ba_status_str[n_i->ba_status]);

        /* report infos */
        _print(cap, info, "      + channel 1    : %s\n",
            n_i->channel_status.summary_ch1 ? "ok" : "nok");
        _print(cap, info, "      + channel 2    : %s\n",
            n_i->channel_status.summary_ch2 ? "ok" : "nok");
    }
    _print(cap, info, "\n");

    _print(cap, info, "  CAPTURED FRAMES    : TOTAL=%d\n", r->stats.frm.total);
    if (r->status.state == FIPCAP_SESS_STATE_RUNNING)
        goto shunt_frm_details;
    _print(cap, info, "    id_dat           : %d\n", r->stats.frm.id_dat);
    _print(cap, info, "    id_msg           : %d\n", r->stats.frm.id_msg);
    _print(cap, info, "    id_rq1           : %d\n", r->stats.frm.id_rq1);
    _print(cap, info, "    id_rq2           : %d\n", r->stats.frm.id_rq2);
    _print(cap, info, "    rp_dat           : %d\n", r->stats.frm.rp_dat);
    _print(cap, info, "    rp_dat_msg       : %d\n", r->stats.frm.rp_dat_msg);
    _print(cap, info, "    rp_dat_rq1       : %d\n", r->stats.frm.rp_dat_rq1);
    _print(cap, info, "    rp_dat_rq2       : %d\n", r->stats.frm.rp_dat_rq2);
    _print(cap, info, "    rp_dat_rq1_msg   : %d\n", r->stats.frm.rp_dat_rq1_msg);
    _print(cap, info, "    rp_dat_rq2_msg   : %d\n", r->stats.frm.rp_dat_rq2_msg);
    _print(cap, info, "    rp_rq1           : %d\n", r->stats.frm.rp_rq1);
    _print(cap, info, "    rp_rq2           : %d\n", r->stats.frm.rp_rq2);
    _print(cap, info, "    rp_msg_noack     : %d\n", r->stats.frm.rp_msg_noack);
    _print(cap, info, "    rp_msg_ack_even  : %d\n", r->stats.frm.rp_msg_ack_even);
    _print(cap, info, "    rp_msg_ack_odd   : %d\n", r->stats.frm.rp_msg_ack_odd);
    _print(cap, info, "    rp_ack_even_p    : %d\n", r->stats.frm.rp_ack_even_p);
    _print(cap, info, "    rp_ack_odd_p     : %d\n", r->stats.frm.rp_ack_odd_p);
    _print(cap, info, "    rp_ack_even_m    : %d\n", r->stats.frm.rp_ack_even_m);
    _print(cap, info, "    rp_ack_odd_m     : %d\n", r->stats.frm.rp_ack_odd_m);
    _print(cap, info, "    rp_fin           : %d\n", r->stats.frm.rp_fin);
    _print(cap, info, "    padding          : %d\n\n", r->stats.frm.padding);
shunt_frm_details:

    _print(cap, info, "  ERRORS DETECTED    : TOTAL=%d\n", r->stats.frm_err.total);
//    if (r->status.state == FIPCAP_SESS_STATE_RUNNING)
//        goto shunt_err_details;
    _print(cap, info, "    ** physical layer errors\n");
    _print(cap, info, "    pre_mis          : %d\n", r->stats.frm_err.pre_mis);
    _print(cap, info, "    fsd_mis          : %d\n", r->stats.frm_err.fsd_mis);
    _print(cap, info, "    fsd_unk          : %d\n", r->stats.frm_err.fsd_unk);
    _print(cap, info, "    fed_mis          : %d\n", r->stats.frm_err.fed_mis);
    _print(cap, info, "    ** data-link layer errors\n");
    _print(cap, info, "    ctl_unk          : %d\n", r->stats.frm_err.ctl_unk);
    _print(cap, info, "    fcs_bad          : %d\n", r->stats.frm_err.fcs_bad);
    _print(cap, info, "    ** application layer errors\n");
    _print(cap, info, "    pdu_bad          : %d\n", r->stats.frm_err.pdu_bad);
    _print(cap, info, "    len_bad          : %d\n", r->stats.frm_err.len_bad);
    _print(cap, info, "    ** library errors\n");
    _print(cap, info, "    idx_loss         : %d\n", r->stats.frm_err.idx_loss);
//shunt_err_details:

    _print(cap, info, "\n");
}

void print_cap_infos(struct fipcap_sess *cap_s)
{
    struct fipcap_sess_infos *inf = cap_s->infos;

    _print(cap, info, "[fip] capture infos\n");
    _print(cap, info, "   sn           : 0x%08x\n",
        inf->sn);
    _print(cap, info,
        "   driver       : v%d.%d.%d [build date: %02d/%02d/%02d]\n",
        inf->drv_version.info.major,
        inf->drv_version.info.minor,
        inf->drv_version.info.patch,
        inf->drv_version.date.info.month,
        inf->drv_version.date.info.day,
        inf->drv_version.date.info.year);
    _print(cap, info,
        "   library      : v%d.%d.%d [build date: %02d/%02d/%02d]\n",
        inf->lib_version.info.major,
        inf->lib_version.info.minor,
        inf->lib_version.info.patch,
        inf->lib_version.date.info.month,
        inf->lib_version.date.info.day,
        inf->lib_version.date.info.year);
}
